==> fxt-2006.06.22.tgz: fxt/src/chirpzt/fftarblen.cc <==

#include "convolution/complexconvolution.h"
#include "chirpzt/chirpzt.h"

#include "aux1/aux1double.h"
#include "aux1/arith1.h"    // negate()
#include "bits/bit2pow.h" // ld()

#include "aux1/copy.h"

#include "fxttypes.h"  // ulong


static void
fft_arblen_even(Complex *x, ulong n, int is)
// Arbitrary length FFT for even lengths n.
{
    ulong ldnn = ld(n);
    if ( n==(1UL<<ldnn) )  ldnn += 1;
    else                   ldnn += 2;
    ulong nn = (1UL<<ldnn);

    Complex *f = new Complex[nn];
    ::copy0(x, n, f, nn);

    Complex *w = new Complex[nn];
    make_fft_chirp(w, n, nn, is);
    multiply(f, n, w);

    double *dw = (double *)w;  // jjcast
    for (ulong k=1; k<2*n; k+=2)  dw[k] = -dw[k];
    // =^=
//    make_fft_chirp(w, n, nn, -is);

    fft_complex_convolution(w, f, ldnn);
    add(f, n, f+n);  // need cyclic convolution

    make_fft_chirp(w, n, nn, is);
    multiply(w, n, f);

    ::copy(w, x, n);
    delete [] w;
    delete [] f;
}
// -------------------------


static void
fft_arblen_odd(Complex *x, ulong n, int is)
// Arbitrary length FFT for odd lengths n.
{
    ulong ldnn = ld(n);
    ldnn += 3;  // (!)
    ulong nn = (1UL<<ldnn);

    Complex *f = new Complex[nn];
    ::copy0(x, n, f, nn);

    Complex *w = new Complex[nn];
    make_fft_chirp_n2(w, n, nn, is);
    multiply(f, n, w);

    double *dw = (double *)w;  // jjcast
    for (ulong k=1; k<nn; k+=2)  dw[k] = -dw[k];
    // =^=
//    make_fft_chirp_n2(w, n, nn, -is);

    fft_complex_convolution(w, f, ldnn);
    ulong n2 = n*2;
    ulong dn = nn-n2;
    if ( dn>n )  dn = n;
    add(f, dn, f+n2);  // need cyclic convolution

    make_fft_chirp(w, n, nn, is);
    multiply(w, n, f);

    ::copy(w, x, n);
    delete [] w;
    delete [] f;
}
// -------------------------


void
fft_arblen(Complex *x, ulong n, int is)
// Arbitrary length FFT.
{
    if ( n&1 )  fft_arblen_odd(x, n, is);
    else        fft_arblen_even(x, n, is);
}
// -------------------------

==> fxt-2006.12.17.tgz: fxt/src/chirpzt/fftarblen.cc <==

#include "convolution/complexconvolution.h"
#include "chirpzt/chirpzt.h"

#include "aux1/aux1double.h"
#include "aux1/arith1.h"    // multiply(), add()
#include "bits/bit2pow.h" // ld()

#include "aux1/copy.h"

#include "fxttypes.h"  // ulong


static void
fft_arblen_even(Complex *x, ulong n, int is)
// Arbitrary length FFT for even lengths n.
{
    ulong ldnn = ld(n);
    if ( n==(1UL<<ldnn) )  ldnn += 1;
    else                   ldnn += 2;
    ulong nn = (1UL<<ldnn);

    Complex *f = new Complex[nn];
    ::copy0(x, n, f, nn);

    Complex *w = new Complex[nn];
    make_fft_chirp(w, n, nn, is);
    multiply(f, n, w);

    double *dw = (double *)w;  // jjcast
    for (ulong k=1; k<2*n; k+=2)  dw[k] = -dw[k];  // =^= make_fft_chirp(w, n, nn, -is);

    fft_complex_convolution(w, f, ldnn);
    add(f, n, f+n);  // need cyclic convolution

    make_fft_chirp(w, n, nn, is);
    multiply(w, n, f);

    ::copy(w, x, n);
    delete [] w;
    delete [] f;
}
// -------------------------


static void
fft_arblen_odd(Complex *x, ulong n, int is)
// Arbitrary length FFT for odd lengths n.
{
    ulong ldnn = ld(n);  // == floor(log2(n))
    ldnn += 3;
    ulong nn = (1UL<<ldnn);  // == 4 * next_2_pow(n)

    Complex *f = new Complex[nn];
    ::copy0(x, n, f, nn);

    Complex *w = new Complex[nn];
    make_fft_chirp_n2(w, n, nn, is);
    multiply(f, n, w);

    double *dw = (double *)w;  // jjcast
    for (ulong k=1; k<nn; k+=2)  dw[k] = -dw[k];  // =^= make_fft_chirp_n2(w, n, nn, -is);

    fft_complex_convolution(w, f, ldnn);
    ulong n2 = n*2;
    ulong dn = nn-n2;
    if ( dn>n )  dn = n;
    add(f, dn, f+n2);  // need cyclic convolution

    make_fft_chirp(w, n, nn, is);
    multiply(w, n, f);

    ::copy(w, x, n);
    delete [] w;
    delete [] f;
}
// -------------------------


void
fft_arblen(Complex *x, ulong n, int is)
// Arbitrary length FFT.
{
    if ( n&1 )  fft_arblen_odd(x, n, is);
    else        fft_arblen_even(x, n, is);
}
// -------------------------
